/**
 * PilrcEdit - a graphical editor for PilRC files.
 * Copyright (C) 2000 Rob Nielsen
 * rnielsen@wn.com.au
 * http://www.wn.com.au/rnielsen/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 */
package pilrcedit;

import java.util.*;
import java.io.*;

public class TextTools
{
  public static final String escapes="\n\t\\";
  public static final String unescapes="nt\\";

  /**
   * Quotes a given string in standard quotes (").  Any appearances of
   * the quote character in the string will be escaped with \"
   * @param s the string to quote
   * @return the quoted string
   */
  public static String quote(String s)
  {
    return quote(s,'"');
  }

  /**
   * Quotes a given string.  Any appearances of
   * the quote character in the string will be escaped.
   * @param s the string to quote
   * @param quote the character to quote with
   * @return the quoted string
   */
  public static String quote(String s,char quote)
  {
    if (s==null||s.equals(""))
      return quote+""+quote;
    return quote+escape(s,""+quote,true)+quote;
  }

  /**
   * Escapes a given string, substituting newlines, tabs and \'s with the
   * strings  "\n","\t", and "\\".  Also, characters less than 32 are replaced
   * with their octal eqivalents (ie \001)
   * @param s the string to escape
   * @return the escaped string
   */
  public static String escape(String s)
  {
		return escape(s,escapes,true);
	}

  /**
   * Escapes a given string, substituting the given characters with their
   * expanded versions.  Also, characters less than 32 or greater than 126
   * can be replaced with their octal eqivalents (ie \001)
   * @param s the string to escape
   * @param escapes the characters to escape
   * @param octal true to escape chars less than 32 or greater than 126, false otherwise
   * @return the escaped string
   */
	public static String escape(String s,String escapes,boolean octal)
	{
		if (s==null)
		  return null;
		//System.out.println("escape(\""+s+"\",\""+escapes+"\","+octal+")");
		boolean changed=false;
		StringBuffer sb=null;
    byte[] bytes=null;
		char[] c=s.toCharArray();
		for(int i=0;i<c.length;i++)
		{
			//System.out.print(c[i]);
			boolean esc=escapes.indexOf(c[i])!=-1;
			if (c[i]=='\\'&&i<c.length-1&&!esc)
			{
        if (escapes.indexOf(c[i+1])!=-1)
        {
					if (sb!=null)
					  sb.append(c[i]).append(c[i+1]);
          i++;
          continue;
				}
			}
			if ((octal&&(c[i]<32||c[i]>126))||esc)
			{
				if (sb==null)
					sb=new StringBuffer(s.substring(0,i));
				changed=true;
				sb.append('\\');
				if (esc)
				{
					if (c[i]=='\n')
					  sb.append('n');
					else
					if (c[i]=='\t')
					  sb.append('t');
					else
					  sb.append(c[i]);
				}
				else
				{
					int oct=c[i];
          if (oct<0||oct>255)
          {
            if (bytes==null)
              bytes=s.getBytes();
            oct=(bytes[i]&0xFF);
          }
					int third=oct%8;
					oct>>=3;
					int second=oct%8;
					oct>>=3;
					int first=oct%8;
	        sb.append(first).append(second).append(third);
				}
			}
			else
			if (sb!=null)
				sb.append(c[i]);
		}
		if (changed)
			return sb.toString();
		return s;
	}

  /**
   * Unescapes a given string, converting literal occurances of
   * "\n", "\t", "\\" and octal values (ie "\001") to their character
   * equivalents.
   * @param s the string to unescape
   * @return the unescaped string
   */
  public static String unescape(String s)
  {
		return unescape(s,unescapes,true);
	}

  /**
   * Unescapes a given string, converting occurances of
   * all characters in unescapes preceded by a \ into normal characters.
   * Optionally octal values (ie "\001") can also be converted.
   * @param s the string to unescape
   * @param unescapes the characters to unescape
   * @param octal true to unescape octal values, false otherwise
   * @return the unescaped string
   */
	public static String unescape(String s,String unescapes,boolean octal)
	{
		if (s==null)
		  return null;
		//System.out.println("unescape(\""+s+"\",\""+escapes+"\","+octal+")");
		boolean changed=false;
		StringBuffer sb=null;
		char[] c=s.toCharArray();
		for(int i=0;i<c.length;i++)
		{
			boolean esc=i<c.length-1&&unescapes.indexOf(c[i+1])!=-1;
			if (c[i]=='\\'&&(esc||(octal&&i<c.length-3)))
			{
				if (sb==null)
  					sb=new StringBuffer(s.substring(0,i));

  			if (esc)
  			{
					char replace=c[i];
					if (c[i+1]=='n')
						replace='\n';
          else
					if (c[i+1]=='t')
						replace='\t';

					sb.append(replace);
					i++;
					continue;
				}
        else //octal
				{
					try
					{
						if (sb==null)
    					sb=new StringBuffer(s.substring(0,i));
						sb.append((char)Integer.parseInt(new String(c,i+1,3),8));
						i+=3;
						continue;
					}
					catch(NumberFormatException e)
					{
					}
				}
			}
			if (sb!=null)
				sb.append(c[i]);
		}
		if (sb!=null)
			return sb.toString();
		return s;
	}

  /**
   * Joins a series of strings together, inserting the given joiner character
   * between them.
   * @param lines the lines to join
   * @param joiner the joiner character
   * @return the joined string
   */
	public static String join(String[] lines,char joiner)
	{
		if (lines==null||lines.length==0)
		  return "";
		StringBuffer sb=new StringBuffer(escape(lines[0],""+joiner,false));
		for(int i=1;i<lines.length;i++)
			sb.append(joiner).append(escape(lines[i],""+joiner,false));
		return sb.toString();
	}

  /**
   * Splits a string into a number of lines on a given character.
   * @param text the text to split
   * @param joiner the character to split on
   * @return an array of the split lines.
   */
	public static String[] split(String text,char joiner)
	{
		Vector list=new Vector();
		String delim="\\"+joiner;
		StringTokenizer st=new StringTokenizer(text,delim,true);
		boolean escape=false;
		boolean inQuote=false;
		StringBuffer sb=null;
		String str=null;
		while(st.hasMoreTokens())
		{
			String s=st.nextToken();
			char ch=s.charAt(0);
			if (ch==joiner)
			{
				if (sb!=null)
  				list.addElement(sb.toString());
  		  else
				if (str!=null)
				  list.addElement(str);
				else
				  list.addElement("");

			  sb=null;
			  str=null;
			  continue;
			}
			if (ch=='\\'&&st.hasMoreTokens())
			{
			  s=st.nextToken();
			  if (s.charAt(0)!=joiner)
			    s="\\"+s;
			}
			if (sb!=null)
			  sb.append(s);
			else
			if (str!=null)
			{
			  sb=new StringBuffer(str);
			  sb.append(s);
			}
			else
			  str=s;
		}
		if (sb!=null)
		  list.addElement(sb.toString());
		else
		if (str!=null)
		  list.addElement(str);
		String[] ret=new String[list.size()];
		for(int i=0;i<ret.length;i++)
			ret[i]=(String)list.elementAt(i);
		return ret;
	}

  /**
   * Splits a string into a number of lines of less than or equal to a given width
   * @param text the text to split
   * @param width the maximum width in characters
   * @return an array of the split lines.
   */
  public static String[] split(String text,int width,boolean preserveNewlines)
  {
    return split(null,text,width,preserveNewlines);
  }

  /**
   * Splits a string into a number of lines of less than or equal to a given
   * pixel width.
   * @param font the font to use.
   * @param text the text to split
   * @param width the maximum width in pixels
   * @return an array of the split lines.
   */
  public static String[] split(Font f,String text,int width)
  {
    return split(f,text,width,false);
  }

  public static String[] split(Font f,String text,int width,boolean preserveNewlines)
  {
		Vector strings=new Vector();
		text=unescape(text);
		StringTokenizer st=new StringTokenizer(text," \n",true);
		String line=null;
		while(st.hasMoreTokens())
		{
			String s=st.nextToken();
			if (s.charAt(0)==' ')
			  continue;
			if (s.charAt(0)=='\n')
			{
        if (preserveNewlines)
          line+='\n';
			  strings.addElement(line);
			  line=null;
			  continue;
			}
			String nline=(line==null?s:line+" "+s);
			if ((f==null&&nline.length()>width)||(f!=null&&f.stringWidth(nline)>width))
			{
				strings.addElement(line);
				line=s;
			}
			else
				line=nline;
		}
		strings.addElement(line);

		String[] ret=new String[strings.size()];
		for(int i=0;i<ret.length;i++)
			ret[i]=(String)strings.elementAt(i);
		return ret;
//		return (String[])strings.toArray(new String[0]);
	}
}